import java.io.BufferedWriter;
import java.io.FileReader;
import java.io.FileWriter;
import java.util.ArrayList;
import java.util.Collections;
import java.util.HashMap;
import java.util.HashSet;
import java.util.Iterator;
import java.util.Map;
import java.util.PriorityQueue;
import java.util.Scanner;
import java.util.Set;

import org.apache.commons.math3.stat.inference.OneWayAnova;
/**
 *
 * @author Joost
 */
public class GA {

	public static void main(String[] args)
	{
		ArrayList<Sample> trainset = new ArrayList<Sample>();
		//Proberen het bestand in te lezen..
		try
		{
			FileReader file = new FileReader("karaktertrain.txt");
			//om een deel van de trainset te lezen:
			int MAXTRAIN = -1;
			int OFFSET = 0;
			Scanner scan = new Scanner( file );
			String line;
			int no = 0;
			System.out.print("Reading trainset.. ");
			for (int i = 0; i < OFFSET && scan.hasNextLine(); i++)
				scan.nextLine();
			while (scan.hasNextLine() && (no < MAXTRAIN || MAXTRAIN < 0)) {
				line = scan.nextLine();
				String[] input = line.split(" ");
				ArrayList<Double> att = new ArrayList<Double>();
				for (int i = 0; i < input.length - 1; i++) {
					att.add(Double.valueOf(input[i]));
				}
				trainset.add(new Sample(att, Integer.valueOf(input[input.length-1]), no++));
			}
			file.close();
			System.out.println("Done!");
		}
		catch ( Exception e )
		{
			System.err.println("Error: " + e.getMessage());
		}
		GA ga = new GA(trainset, new ModifiedRK_GA());
		createResults(ga.findFittestClassifier(), trainset);
		System.out.println("Classification finished.\nExiting..");
	}

	private PriorityQueue<Specimen> population;
	private GeneticStrategy strategy;
	private static final int ITERATIONS = 100;
	public static final int POPSIZE = 10;
	private static final double SIGNIFICANCE_LEVEL = 0.000005;
	
	public GA(ArrayList<Sample> trainset, GeneticStrategy strategy) {
		this.strategy = strategy;
		population = new PriorityQueue<Specimen>();
		System.out.println("Constructing");
		HashSet<Integer> filteredAttr = filterAttributeSpread(trainset);
		System.out.println("Removed "+filteredAttr.size()+" attributes based on ANOVA: "+filteredAttr);
		for (int i = 0; i < POPSIZE; i++) {
			System.out.println(i);
			Specimen s = new Specimen(trainset, filteredAttr);
			s.setFitness(strategy.fitness(s));
			population.add(s);
		}
		System.out.println("Done constructing");
	}
	
	private HashSet<Integer> filterAttributeSpread(ArrayList<Sample> trainset) {
		HashSet<Integer> h = new HashSet<Integer>();
		//Omdat we alleen een ANOVA-test kunnen doen wanneer er evenveel waarden bekend zijn per klasse
		//Zullen we eerst even moeten tellen hoeveel samples er minimaal bestaan voor elke klasse
		int minCountPerClass = 0;
		HashMap<Integer, Integer> classSizeMap = new HashMap<Integer, Integer>();
		for (Sample s : trainset) {
			if (classSizeMap.get(s.getClassification()) == null)
				classSizeMap.put(s.getClassification(), 1);
			else
				classSizeMap.put(s.getClassification(), classSizeMap.get(s.getClassification()) + 1);
		}
		//Klassen waar er maar 1 van is kunnen weg, want daar kan ANOVA toch niks mee
		Set<Map.Entry<Integer,Integer>> entrySet = classSizeMap.entrySet();
		ArrayList<Integer> removeables = new ArrayList<Integer>();
		for (Map.Entry<Integer,Integer> map : entrySet) {
			if (map.getValue() == 1) {
				removeables.add(map.getKey());
			}
		}
		for (int n : removeables) {
			classSizeMap.remove(n);
		}
		minCountPerClass = Collections.min(classSizeMap.values());
		//Daarna gaan we alle attributen af, en kijken we of ze aan de ANOVA-test voldoen
		for (int i = 0; i < trainset.get(0).getAttributeCount(); i++)
		{
			HashMap<Integer, Integer> indexMap = new HashMap<Integer, Integer>();
			HashMap<Integer, Integer> countMap = new HashMap<Integer, Integer>();
			ArrayList<double[]> classes = new ArrayList<double[]>();
			for (Sample s : trainset) {
				if (indexMap.get(s.getClassification()) == null) {
					double[] d = new double[minCountPerClass];
					indexMap.put(s.getClassification(), classes.size());
					countMap.put(s.getClassification(), 0);
					classes.add(d);
				}
				if (minCountPerClass > countMap.get(s.getClassification())) {
					double[] values = classes.get(indexMap.get(s.getClassification()));
					values[countMap.get(s.getClassification())] = s.getAttributeValue(i);
					countMap.put(s.getClassification(), countMap.get(s.getClassification()) + 1);
				}
			}
			OneWayAnova anova = new OneWayAnova();
			if (!anova.anovaTest(classes, SIGNIFICANCE_LEVEL)) {
				h.add(i);
			}
		}
		return h;
	}

	private Specimen findFittestClassifier() {
		System.out.println("Start iterating..");
		for (int i = 0; i < ITERATIONS; i++) {
			strategy.iterate(population);
			System.out.println(i);
		}
		System.out.println("Done iterating");
		Specimen s = strategy.getFittestSpecimen(population);
		System.out.println("The fittest classifier is:\n"+s);
		System.out.println("With a fitness of: "+strategy.fitness(s));
		return s;
	}
	
	private static void createResults(Specimen classifier, ArrayList<Sample> trainset) {
		try {
			System.out.print("Start creating prototypes.. ");
			// Maken van prototype-file
			FileWriter ofstream = new FileWriter("karakterprototypes.txt");
			BufferedWriter out = new BufferedWriter(ofstream);
			for (int e : classifier.getExamples()) {
				Sample s = trainset.get(e);
				Iterator<Integer> iteratorA = classifier.getAttributes().iterator();
				int a = iteratorA.hasNext() ? iteratorA.next() : -1;
				for (int i = 0; i < s.getAttributeCount(); i++) {
					if (a > i) {
						continue;
					}
					else if (a == i) {
						out.write(s.getAttributeValue(i)+" ");
						a = iteratorA.hasNext() ? iteratorA.next() : -1;
					}
				}
				out.write(s.getClassification()+"\n");
			}
			//Sluiten van schrijven naar prototype-file
			out.close();
			System.out.println("Done!");
		}
		catch (Exception e) {
			System.err.println("Error 1: " + e.getMessage());
		}
		try {
			//Openen van de testset
			FileReader ifstream = new FileReader("karaktertest.txt");
			Scanner scan = new Scanner( ifstream );
			//Maken van een predictionbestand
			FileWriter ofstream = new FileWriter("karakterpredictions.txt");
			BufferedWriter out = new BufferedWriter(ofstream);
						
			String line;
			System.out.print("Start predicting testset.. ");
			while (scan.hasNextLine()) {
				line = scan.nextLine();
				String[] input = line.split(" ");
				ArrayList<Double> att = new ArrayList<Double>();
				for (int i = 0; i < input.length; i++) {
					att.add(Double.valueOf(input[i]));
				}
				out.write(classifier.computeClassification(att)+"\n");
			}
			
			out.close();
			ifstream.close();
			System.out.println("Done!");
		}
		catch (Exception e) {
			System.err.println("Error 2: " + e.getMessage());
		}
	}
}
