import java.util.ArrayList;
import java.util.Collections;
import java.util.Comparator;
import java.util.Iterator;
import java.util.PriorityQueue;
import java.util.Random;

/**
 *
 * @author Joost
 */
public class RK_GA implements GeneticStrategy {

	private double c1, c2, c3;
	protected int MATINGPAIRS;
	protected final double MUTATIONRATE;
	
	public RK_GA() {
		c1 = 1; //correctness
		c2 = 1; //examplecount
		c3 = 1; //attributecount
		MATINGPAIRS = 10;
		MUTATIONRATE = 0.02;
	}
	
	public double fitness(Specimen s) {
		int Er = s.getIncorrectSamples();
		int Ne = s.getExamplesCount();
		int Na = s.getAttributeCount();
		return 1 / (c1 * Er + c2 * Ne + c3 * Na);
	}

	public void iterate(PriorityQueue<Specimen> population) {
		if (MATINGPAIRS > GA.POPSIZE/2)
			MATINGPAIRS = GA.POPSIZE/2;
		for (int i = 0; i < MATINGPAIRS; i++) {
			Specimen[] specs = selectSpecimen(population);
			Specimen s1 = specs[0];
			Specimen s2 = specs[1];
			specs = twoPointCrossOver(s1, s2);
			specs[0] = mutate(specs[0]);
			specs[0].setFitness(fitness(specs[0]));
			specs[1] = mutate(specs[1]);
			specs[1].setFitness(fitness(specs[1]));
			population.add(specs[0]);
			population.add(specs[1]);
		}
	}
	
	protected Specimen[] twoPointCrossOver(Specimen s1, Specimen s2) {
		Specimen[] specs = new Specimen[2];
		Random r = new Random();
		//Deze willekeurige getallen geven de bounds van substrings aan binnen de chromosomen van s1 en s2
		int r1a1 = 0, r1a2 = 0, r2a1 = 0, r2a2 = 0, r1e1 = 0, r1e2 = 0, r2e1 = 0, r2e2 = 0;
		if (s1.getAttributeCount() > 0) {
			r1a1 = r.nextInt(s1.getAttributeCount());
			r1a2 = r.nextInt(s1.getAttributeCount());
		}
		if (s1.getExamplesCount() > 0) {
			r1e1 = r.nextInt(s1.getExamplesCount());
			r1e2 = r.nextInt(s1.getExamplesCount());
		}
		if (s2.getAttributeCount() > 0) {
			r2a1 = r.nextInt(s2.getAttributeCount());
			r2a2 = r.nextInt(s2.getAttributeCount());
		}
		if (s2.getExamplesCount() > 0) {
			r2e1 = r.nextInt(s2.getExamplesCount());
			r2e2 = r.nextInt(s2.getExamplesCount());
		}
		ArrayList<Integer> A_att = twoPointCrossover(s1.getAttributes(), s2.getAttributes(), Math.min(r1a1, r1a2), Math.max(r1a1, r1a2), Math.min(r2a1, r2a2), Math.max(r2a1, r2a2));
		ArrayList<Integer> B_att = twoPointCrossover(s2.getAttributes(), s1.getAttributes(), Math.min(r2a1, r2a2), Math.max(r2a1, r2a2), Math.min(r1a1, r1a2), Math.max(r1a1, r1a2));
		ArrayList<Integer> A_exmp = twoPointCrossover(s1.getExamples(), s2.getExamples(), Math.min(r1e1, r1e2), Math.max(r1e1, r1e2), Math.min(r2e1, r2e2), Math.max(r2e1, r2e2));
		ArrayList<Integer> B_exmp = twoPointCrossover(s2.getExamples(), s1.getExamples(), Math.min(r2e1, r2e2), Math.max(r2e1, r2e2), Math.min(r1e1, r1e2), Math.max(r1e1, r1e2));
		specs[0] = new Specimen(A_att, A_exmp, s1.getSamples());
		specs[1] = new Specimen(B_att, B_exmp, s1.getSamples());
		
		return specs;
	}

	protected ArrayList<Integer> twoPointCrossover(ArrayList<Integer> a, ArrayList<Integer> b, int a1, int a2, int b1, int b2) {
		ArrayList<Integer> r = new ArrayList<Integer>();
		r.addAll(a.subList(0, a1));
		r.addAll(b.subList(b1, b2));
		r.addAll(a.subList(a2, a.size()));
		sortAndUnique(r);
		return r;
	}
	
	public Specimen getFittestSpecimen(PriorityQueue<Specimen> population) {
		return population.peek();
	}

	protected class FitnessComparator implements Comparator<Specimen> {

		public int compare(Specimen o1, Specimen o2) {
			if (fitness(o1) < fitness(o2))
				return -1;
			return (fitness(o1) == fitness(o2) ? 0 : 1);
		}
		
	}
	
	protected Specimen mutate(Specimen s) {
		mutateChromosome(s.getAttributes(), s.getSampleAttributeCount());
		mutateChromosome(s.getExamples(), s.getSampleCount());
		return s;
	}
	
	protected void mutateChromosome(ArrayList<Integer> c, int total) {
		Random r = new Random();
		for (int i = 0; i < Math.ceil(c.size() * MUTATIONRATE); i++) {
			int r2 = r.nextInt(c.size());
			int n = r.nextInt(c.size());
			c.set(r2, (c.get(r2) + n) % total);
		}
		sortAndUnique(c);
	}
	
	protected void sortAndUnique(ArrayList<Integer> c) {
		Collections.sort(c);
		Iterator<Integer> it = c.iterator();
		int prev = -1, curr = -1;
		while (it.hasNext()) {
			curr = it.next();
			if (curr == prev) {
				it.remove();
			}
			prev = curr;
		}
	}
	
	protected Specimen[] selectSpecimen(PriorityQueue<Specimen> population) {
		Specimen[] specs = new Specimen[2];
		specs[0] = selectRandomSpecimen(population);
		specs[1] = selectRandomSpecimen(population);
		return specs;
	}
	
	/**
	 * Deze functie levert een willekeurige Specimen op uit de populatie,
	 * volgens de niet-uniforme distributie van fitness
	 * @param population de populatie van specimen
	 * @return een willekeurige specimen
	 */
	protected Specimen selectRandomSpecimen(PriorityQueue<Specimen> population) {
		double sumOfFitness = 0;
		int i = 0;
		Iterator<Specimen> it = population.iterator();
		while (i < GA.POPSIZE && it.hasNext()) {
			sumOfFitness += fitness(it.next());
			i++;
		}
		it = population.iterator();
		Random r = new Random();
		double rd = r.nextDouble();
		double n = 0;
		i = -1;
		Specimen result = it.next();
		do {
			i++;
			n += fitness(result) / sumOfFitness;
			if (it.hasNext())
				result = it.next();
		} while (n < rd);
		return result;
	}
}
